/* ------------------------------------------------------------------------------------------------ */
/* FILE :		 Demo_2_MSP5438TK_CCS.c																*/
/*																									*/
/* CONTENTS :																						*/
/*				Demo for the Olimex MSP-5438STK evaluation board									*/
/*				Operation of the RTOS shown on the LCD display										*/
/*				Control of the tasks parameters through both serial ports (RS232 & RS232 over USB)	*/
/*				Serial port settings:																*/
/*							Baud Rate: 115200														*/
/*							Data bits:      8														*/
/*							Stop bits:      1														*/
/*							Parity:      none														*/
/*							Flow Ctrl:   none														*/
/*							Emulation: VT-100														*/
/*																									*/
/*																									*/
/* Copyright (c) 2012-2014, Code-Time Technologies Inc. All rights reserved.						*/
/*																									*/
/* Code-Time Technologies retains all right, title, and interest in and to this work				*/
/*																									*/
/* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS							*/
/* OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF										*/
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL							*/
/* THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR								*/
/* OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,							*/
/* ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR							*/
/* OTHER DEALINGS IN THE SOFTWARE.																	*/
/*																									*/
/*																									*/
/*	$Revision: 1.2 $																				*/
/*	$Date: 2012/06/17 15:00:29 $																	*/
/*																									*/
/* ------------------------------------------------------------------------------------------------ */

#if OS_DEMO != 2
	#error "Wrong OS_DEMO value" 
#endif

#include <stdio.h>
#include <stdarg.h>
#include <ctype.h>
#include "Abassi.h"

/* ------------------------------------------------ */

#include <msp430.h>
#include <file.h>

#define BAUDRATE	  (115200L)						/* Serial port baud rate						*/
#define LCD_FREQ	 (1000000L)						/* LCD serial port bit rate						*/
#define SYS_FREQ	 (9000000L)						/* Processor clock frequency					*/
#define OSC_FREQ	(18000000L)						/* Oscillator clock frequency					*/

void CLOCK_XT2_Init(void);
void Delay(unsigned long time);
void LCD_Clear();
void LCD_Clear_Memory();
void LCD_Contrast(int contrast);
void LCD_Init();
void LCD_Send(int data, int cmd);
void LCD_Update(int mode, int num, int y);
void LCD_Write(char *a, int x, int y);
void SPI2_Init();
void TIMERinit(long Rate);
void UARTinit(void);
void UART2_Init(unsigned long baud);
void UART3_Init(unsigned long baud);


#define SEND_CMD 			0						/* LCD commands									*/
#define SEND_CHR 			1
#define LCD_START_LINE_ADDR	(66-2)

int __putchar(int ccc);
int __getchar(void);

/* ------------------------------------------------ */

#define TSK_REFRESH_RATE	5						/* Display refresh rate: 1/TSK_REFRESH_RATE sec	*/
#define RTC_REFRESH_RATE	1						/* RTC refresh rate related to TSK_REFRESH_RATE	*/

#define USE_EVENTS			0						/* Events or mailboxes used to inform task 		*/

#define OP_ZERO_CNT			1						/* Operation requested by the ISR state machine	*/
#define OP_LOCK_MTX			2
#define OP_UNLOCK_MTX		4

#define TSK_HIGH			0						/* Index of the task info in TskDsc[]			*/
#define TSK_RRA				1
#define TSK_RRB				2
#define TSK_RRC				3
#define TSK_STRVA			4
#define TSK_STRVB			5

#define OP_ROUND_ROBIN		0						/* Command line configuration operation			*/
#define OP_STARVE_RUN		1
#define OP_STARVE_WAIT		2
#define OP_STARVE_PRIO		3
#define OP_PRIO				4

#define MBX_DISP_SIZE		32						/* Depth of the mailbox for the display			*/

/* ------------------------------------------------ */

typedef struct {									/* Hold the descriptions for the task monitored	*/
	TSK_t *Task;									/* Task descriptor								*/
  #if ((USE_EVENTS) == 0)
	MBX_t *Mailbox;									/* Mailbox when ISR state machine uses mailboxes*/
  #endif
	int    Running;									/* If this is the running task (ignoring A&E	*/
	int    WasReset;								/* To inform task to reset running time counter	*/
	int    CPUcnt;									/* Number of timer ticks task has been running	*/
	int    WillBlk;									/* If the task is next to block on the mutex	*/
	int    WillMtx;									/* If the task is next to lock the mutex		*/
} TskDsc_t;
TskDsc_t TskDsc[6];

typedef struct {									/* Info sent to the task to update the display	*/
	char *Str;										/* String to print on the display				*/
	int  X;											/* X coordinate									*/
	int  Y;											/* Y coordinate									*/
} DispInfo_t;
DispInfo_t DispInfo[MBX_DISP_SIZE];

volatile int g_DispIdx=0;							/* Entry in DispInfo[] is next to be filled		*/

char Display[7][20] = { "00.0 Px    ",
                        "00.0 Px    ",
                        "00.0 Px    ",
                        "00.0 Px    ",
                        "00.0 Px    ",
                        "00.0 Px    ",
						"00:00:00   "
                      };
volatile int g_CmdIdx;								/* Index where to write next character from kyb	*/
TSK_t *AdamEve;
MTX_t *Mutex;										/* Mutex used by the task to show inheritance	*/
MBX_t *MbxDisp;										/* Mailbox to send display updates to the tasks	*/
SEM_t *SemaAE;										/* Semaphore to awake Adam & Eve				*/

#if ((OX_STARVE_RUN_MAX) != 0)
  extern TSK_t *g_StrvList;							/* Imported from Abassi.c						*/
#else
  TSK_t *g_StrvList = (TSK_t *)NULL;
#endif

/* ------------------------------------------------ */

void   Fct(void);									/* A single function implements all the tasks	*/
TSK_t *GetHeadPrio(int Prio);						/* To obtain ready to run at desired priority	*/
int    GetKey(void);								/* Non-blocking keyboard input					*/

/* ------------------------------------------------ */

static const char FontLookup [][5] =		/* 5X7 fonts for the LCD display				*/
{
    { 0x00, 0x00, 0x00, 0x00, 0x00 },   // sp
    { 0x00, 0x00, 0x2f, 0x00, 0x00 },   // !
    { 0x00, 0x07, 0x00, 0x07, 0x00 },   // "
    { 0x14, 0x7f, 0x14, 0x7f, 0x14 },   // #
    { 0x24, 0x2a, 0x7f, 0x2a, 0x12 },   // $
    { 0xc4, 0xc8, 0x10, 0x26, 0x46 },   // %
    { 0x36, 0x49, 0x55, 0x22, 0x50 },   // &
    { 0x00, 0x05, 0x03, 0x00, 0x00 },   // '
    { 0x00, 0x1c, 0x22, 0x41, 0x00 },   // (
    { 0x00, 0x41, 0x22, 0x1c, 0x00 },   // )
    { 0x14, 0x08, 0x3E, 0x08, 0x14 },   // *
    { 0x08, 0x08, 0x3E, 0x08, 0x08 },   // +
    { 0x00, 0x00, 0x50, 0x30, 0x00 },   // ,
    { 0x10, 0x10, 0x10, 0x10, 0x10 },   // -
    { 0x00, 0x60, 0x60, 0x00, 0x00 },   // .
    { 0x20, 0x10, 0x08, 0x04, 0x02 },   // /
    { 0x3E, 0x51, 0x49, 0x45, 0x3E },   // 0
    { 0x00, 0x42, 0x7F, 0x40, 0x00 },   // 1
    { 0x42, 0x61, 0x51, 0x49, 0x46 },   // 2
    { 0x21, 0x41, 0x45, 0x4B, 0x31 },   // 3
    { 0x18, 0x14, 0x12, 0x7F, 0x10 },   // 4
    { 0x27, 0x45, 0x45, 0x45, 0x39 },   // 5
    { 0x3C, 0x4A, 0x49, 0x49, 0x30 },   // 6
    { 0x01, 0x71, 0x09, 0x05, 0x03 },   // 7
    { 0x36, 0x49, 0x49, 0x49, 0x36 },   // 8
    { 0x06, 0x49, 0x49, 0x29, 0x1E },   // 9
    { 0x00, 0x36, 0x36, 0x00, 0x00 },   // :
    { 0x00, 0x56, 0x36, 0x00, 0x00 },   // ;
    { 0x08, 0x14, 0x22, 0x41, 0x00 },   // <
    { 0x14, 0x14, 0x14, 0x14, 0x14 },   // =
    { 0x00, 0x41, 0x22, 0x14, 0x08 },   // >
    { 0x02, 0x01, 0x51, 0x09, 0x06 },   // ?
    { 0x32, 0x49, 0x59, 0x51, 0x3E },   // @
    { 0x7E, 0x11, 0x11, 0x11, 0x7E },   // A
    { 0x7F, 0x49, 0x49, 0x49, 0x36 },   // B
    { 0x3E, 0x41, 0x41, 0x41, 0x22 },   // C
    { 0x7F, 0x41, 0x41, 0x22, 0x1C },   // D
    { 0x7F, 0x49, 0x49, 0x49, 0x41 },   // E
    { 0x7F, 0x09, 0x09, 0x09, 0x01 },   // F
    { 0x3E, 0x41, 0x49, 0x49, 0x7A },   // G
    { 0x7F, 0x08, 0x08, 0x08, 0x7F },   // H
    { 0x00, 0x41, 0x7F, 0x41, 0x00 },   // I
    { 0x20, 0x40, 0x41, 0x3F, 0x01 },   // J
    { 0x7F, 0x08, 0x14, 0x22, 0x41 },   // K
    { 0x7F, 0x40, 0x40, 0x40, 0x40 },   // L
    { 0x7F, 0x02, 0x0C, 0x02, 0x7F },   // M
    { 0x7F, 0x04, 0x08, 0x10, 0x7F },   // N
    { 0x3E, 0x41, 0x41, 0x41, 0x3E },   // O
    { 0x7F, 0x09, 0x09, 0x09, 0x06 },   // P
    { 0x3E, 0x41, 0x51, 0x21, 0x5E },   // Q
    { 0x7F, 0x09, 0x19, 0x29, 0x46 },   // R
    { 0x46, 0x49, 0x49, 0x49, 0x31 },   // S
    { 0x01, 0x01, 0x7F, 0x01, 0x01 },   // T
    { 0x3F, 0x40, 0x40, 0x40, 0x3F },   // U
    { 0x1F, 0x20, 0x40, 0x20, 0x1F },   // V
    { 0x3F, 0x40, 0x38, 0x40, 0x3F },   // W
    { 0x63, 0x14, 0x08, 0x14, 0x63 },   // X
    { 0x07, 0x08, 0x70, 0x08, 0x07 },   // Y
    { 0x61, 0x51, 0x49, 0x45, 0x43 },   // Z
    { 0x00, 0x7F, 0x41, 0x41, 0x00 },   // [
    { 0x55, 0x2A, 0x55, 0x2A, 0x55 },   // \ .
    { 0x00, 0x41, 0x41, 0x7F, 0x00 },   // ]
    { 0x04, 0x02, 0x01, 0x02, 0x04 },   // ^
    { 0x40, 0x40, 0x40, 0x40, 0x40 },   // _
    { 0x00, 0x01, 0x02, 0x04, 0x00 },   // '
    { 0x20, 0x54, 0x54, 0x54, 0x78 },   // a
    { 0x7F, 0x48, 0x44, 0x44, 0x38 },   // b
    { 0x38, 0x44, 0x44, 0x44, 0x20 },   // c
    { 0x38, 0x44, 0x44, 0x48, 0x7F },   // d
    { 0x38, 0x54, 0x54, 0x54, 0x18 },   // e
    { 0x08, 0x7E, 0x09, 0x01, 0x02 },   // f
    { 0x0C, 0x52, 0x52, 0x52, 0x3E },   // g
    { 0x7F, 0x08, 0x04, 0x04, 0x78 },   // h
    { 0x00, 0x44, 0x7D, 0x40, 0x00 },   // i
    { 0x20, 0x40, 0x44, 0x3D, 0x00 },   // j
    { 0x7F, 0x10, 0x28, 0x44, 0x00 },   // k
    { 0x00, 0x41, 0x7F, 0x40, 0x00 },   // l
    { 0x7C, 0x04, 0x18, 0x04, 0x78 },   // m
    { 0x7C, 0x08, 0x04, 0x04, 0x78 },   // n
    { 0x38, 0x44, 0x44, 0x44, 0x38 },   // o
    { 0x7C, 0x14, 0x14, 0x14, 0x08 },   // p
    { 0x08, 0x14, 0x14, 0x18, 0x7C },   // q
    { 0x7C, 0x08, 0x04, 0x04, 0x08 },   // r
    { 0x48, 0x54, 0x54, 0x54, 0x20 },   // s
    { 0x04, 0x3F, 0x44, 0x40, 0x20 },   // t
    { 0x3C, 0x40, 0x40, 0x20, 0x7C },   // u
    { 0x1C, 0x20, 0x40, 0x20, 0x1C },   // v
    { 0x3C, 0x40, 0x30, 0x40, 0x3C },   // w
    { 0x44, 0x28, 0x10, 0x28, 0x44 },   // x
    { 0x0C, 0x50, 0x50, 0x50, 0x3C },   // y
    { 0x44, 0x64, 0x54, 0x4C, 0x44 },   // z
    { 0x08, 0x6C, 0x6A, 0x19, 0x08 },   // {
    { 0x30, 0x4E, 0x61, 0x4E, 0x30 }    // |
};

char LCDMemory[504];
int  g_LCDidx;

													/* LCD display refresh so slow, used this to 	*/
													/* update it only when needed					*/
#define SET_DISP(_idx, _cc) do {if (String[_idx] != _cc) {Update=1; String[_idx]=_cc;}} while(0)

/* ------------------------------------------------------------------------------------------------ */
/* ------------------------------------------------------------------------------------------------ */

int main(void)
{
int   ccc;											/* Character read from the keyboard				*/
int   Done;											/* If command decoding is done					*/
int   Error;										/* If error during command decoding				*/
int   ii;											/* General purpose								*/
int   jj;											/* General purpose								*/
int   kk;											/* General purpose								*/
int   Op;											/* Operation requested by the command			*/
int   RdIdx;										/* Read index in the command line string		*/
char *String;										/* Display sting being created					*/
int   TaskIdx;										/* Index of the task specified on command line	*/
int   Toggle;										/* Toggle use to control refresh of RTC			*/
int   Update;										/* If the line on the LCD should be updated		*/

static char Cmd[64];								/* Command line from keyboard					*/

/* ------------------------------------------------ */
/* 	Processor clock set-up							*/

	WDTCTL = WDTPW + WDTHOLD;						/* Turn off the watchdog timer					*/
													/* Done earlier in IAR with low_lelevl_init()	*/
	CLOCK_XT2_Init();

/* ------------------------------------------------ */
/* 	Start the RTOS									*/

	OSstart();

	AdamEve = TSKmyID();

/* ------------------------------------------------ */
/* UART set-up										*/

	UARTinit();

/* ------------------------------------------------ */
/* Application set-up								*/

													/* Need to create all services before allowing	*/
	MbxDisp                  = MBXopen("Disp", MBX_DISP_SIZE);	/* any task to run					*/
	Mutex                    = MTXopen("Mutex");
	SemaAE                   = SEMopen("Sem A&E"); 

	TskDsc[TSK_HIGH].Task    = TSKcreate("HI   ", 2, TC_TSK_STACK, &Fct, 0);
	TskDsc[TSK_RRA].Task     = TSKcreate("RRA  ", 4, TC_TSK_STACK, &Fct, 0);
	TskDsc[TSK_RRB].Task     = TSKcreate("RRB  ", 4, TC_TSK_STACK, &Fct, 0);
	TskDsc[TSK_RRC].Task     = TSKcreate("RRC  ", 4, TC_TSK_STACK, &Fct, 0);

	TskDsc[TSK_STRVA].Task   = TSKcreate("STRVA", 7, TC_TSK_STACK, &Fct, 0);
	TskDsc[TSK_STRVB].Task   = TSKcreate("STRVB", 9, TC_TSK_STACK, &Fct, 0);

  #if ((OX_ROUND_ROBIN) < 0)
	TSKsetRR(TskDsc[TSK_RRA].Task, 4*OX_RR_TICK);
	TSKsetRR(TskDsc[TSK_RRB].Task, 6*OX_RR_TICK);
	TSKsetRR(TskDsc[TSK_RRC].Task, 8*OX_RR_TICK);
  #endif

  #if ((OX_STARVE_RUN_MAX) < 0)
	TSKsetStrvRunMax (TskDsc[TSK_STRVA].Task,   5*OS_TICK_PER_SEC);
	TSKsetStrvRunMax (TskDsc[TSK_STRVB].Task,   5*OS_TICK_PER_SEC);
  #endif

  #if ((OX_STARVE_WAIT_MAX) < 0)
	TSKsetStrvWaitMax(TskDsc[TSK_STRVA].Task, (25*OS_TICK_PER_SEC)/10);
	TSKsetStrvWaitMax(TskDsc[TSK_STRVB].Task, (10*OS_TICK_PER_SEC)/10);
  #endif

	for (ii=0 ; ii<sizeof(TskDsc)/sizeof(TskDsc[0]) ; ii++) {
		TSKsetArg(TskDsc[ii].Task, &TskDsc[ii]);
	  #if ((USE_EVENTS) == 0)
		TskDsc[ii].Mailbox  = MBXopen(NULL, 32);
	  #endif
		TskDsc[ii].CPUcnt   = 0;
		TskDsc[ii].Running  = 0;
		TskDsc[ii].WasReset = 0;
		TskDsc[ii].WillBlk  = 0;
		TskDsc[ii].WillMtx  = 0;
		TSKresume(TskDsc[ii].Task);
	}

/* ------------------------------------------------ */
/* LCD set-up										*/

	SPI2_Init();
	LCD_Init();

	g_LCDidx = 0;

	LCD_Contrast(255);

	LCD_Write("Hi 00.0 Px", 0, 0);
	LCD_Write("RA 00.0 Px", 0, 1);
	LCD_Write("RB 00.0 Px", 0, 2);
	LCD_Write("RC 00.0 Px", 0, 3);
	LCD_Write("SA 00.0 Px", 0, 4);
	LCD_Write("SB 00.0 Px", 0, 5);
	for (ii=0 ; ii<6 ; ii++) {
	  	LCD_Update(1, 0, ii);
	}

/* ------------------------------------------------ */
/* SysTick Set-up									*/

	TIMERinit(((SYS_FREQ/100000L)*((long)OS_TIMER_US))/10L);
	OSisrInstall(TIMER0_A0_VECTOR, &TIMtick);

/* ------------------------------------------------ */
/* Interrupts ON, the application runs				*/

	OSeint(1);

	g_CmdIdx = 0;
	Toggle   = 0;

	printf("\033[15;0H");							/* Set cursor where desired						*/
	printf("Abassi Demo > ");						/* Prompt on Serial for the command line		*/

	for (;;) {

		SEMwait(SemaAE, -1);

		if (++Toggle >= RTC_REFRESH_RATE) { 		/* RTC is not always refreshed					*/
			Toggle = 0;
		}

		for (ii=0 ; ii<sizeof(TskDsc)/sizeof(TskDsc[0]) ; ii++) {	/* Update all tasks status		*/
			String = &Display[ii][0];				/* Display line to update						*/
			Update = 0;

			jj = TskDsc[ii].CPUcnt;
			if (jj > 9999) {						/* Can't display more than 99.9 (is /10)		*/
				jj = 9999;
			}
			kk = jj/1000;							/* Running time MS digit						*/
			SET_DISP(0, '0'+kk);
			jj = jj - (kk*1000);					/* Running time LS digit						*/
			kk = jj/100;
			SET_DISP(1, '0'+kk);
			jj = jj - (kk*100);						/* Running time fraction						*/
			kk = jj/10;
			SET_DISP(3, '0'+kk);
			if (TSKisSusp(TskDsc[ii].Task)) {		/* If the task is suspended, the priority		*/
				SET_DISP(5, 'S');
				SET_DISP(6, ' ');
			}
			else {									/* Regular priority field						*/
				SET_DISP(5, 'P');
				if (TskDsc[ii].Task->SuspRqst != 0) {
					SET_DISP(6, 'S');				/* When the task will be suspended, indicate so	*/
				}
				else {								/* Priority numerical value						*/
					SET_DISP(6, '0'+TSKgetPrio(TskDsc[ii].Task));
				}
			}
			if (TskDsc[ii].Running) {				/* If this is the running task, indicate so		*/
				SET_DISP(8, 'R');
			}
			else {									/* Not the running task, nothing in this field	*/
				SET_DISP(8, ' ');
			}

			if (TskDsc[ii].Task == g_StrvList) {	/* If the task is under starvation protection	*/
				SET_DISP(9, 'A');					/* Show the "A"									*/
			}
			else {									/* Not under starvation protection, nothing to	*/
				SET_DISP(9, ' ');					/* shown in this field							*/
			}

			if (TskDsc[ii].Task->MyMutex == Mutex) {
				SET_DISP(10, 'M');					/* If this task locks the mutex, show it		*/
			}
			else if (TskDsc[ii].WillMtx  != 0) {
				SET_DISP(10, 'm');					/* If the task will lock the mutex, show it		*/
			}
			else if (TskDsc[ii].Task->Blocker == Mutex) {
				SET_DISP(10, 'B');					/* If the task is blocked by the mutex, show it	*/
			}
			else if (TskDsc[ii].WillBlk  != 0) {
				SET_DISP(10, 'b');					/* if the task will block on the mutex, show it	*/
			}
			else {
				SET_DISP(10, ' ');					/* If the task not involved with the mutex, 	*/
			}										/* nothing to show in this field				*/

			if ((TSKgetPrio(TskDsc[ii].Task) != TskDsc[ii].Task->OriPrio)
			&&  ((TskDsc[ii].Task != g_StrvList)
			  || (TskDsc[ii].Task->OriPrio != TskDsc[ii].Task->StrvPrio))
			&&  (TskDsc[ii].Task->MyMutex == Mutex)) {
				SET_DISP(10, 'I');					/* If the task has it's priority increased due	*/
			}										/* to priority inheritance, show this			*/
			if (Update != 0) {
				DispInfo[g_DispIdx].Str = String;
				DispInfo[g_DispIdx].X   = ii;
				DispInfo[g_DispIdx].Y   = 16;
													/* Send the string to display to running task	*/
				MBXput(MbxDisp, (intptr_t)&DispInfo[g_DispIdx], 0);
				if (++g_DispIdx >= (sizeof(DispInfo)/sizeof(DispInfo[0]))) {
					g_DispIdx = 0;
				}
			}
		}

		ccc = GetKey();								/* Check for a new key pressed on the keyboard 	*/
		if (ccc != 0) {
			if (g_CmdIdx == 0) {
				for (ii=0 ; ii<sizeof(Cmd) ; ii++) {
					putchar(' ');
				}
				printf("\033[15;15H");
			}
			if (g_CmdIdx < (sizeof(Cmd)/sizeof(Cmd[0]))-2) {
				putchar(ccc);						/* A key was pressed , echo it					*/
			}										/* If command line buffer full, don't echo		*/
			if (ccc == '\r') {						/* Need to remap CR to '\n'						*/
				putchar('\n');
			}
			else if (ccc == '\b') {					/* Backspace pressed, to erase previous key		*/
				if (g_CmdIdx != 0) {
					g_CmdIdx--;
					putchar(' ');
					putchar('\b');
				}
			}
			else {
				if (g_CmdIdx == 0) {				/* Add to the command line without overflow		*/
					printf("\033[15;16H");
					for (ii=1 ; ii<sizeof(Cmd) ; ii++) {
						putchar(' ');
					}
					printf("\033[15;16H");
				}
				if (g_CmdIdx < (sizeof(Cmd)/sizeof(Cmd[0]))-2) {
					Cmd[g_CmdIdx] = toupper(ccc);
					g_CmdIdx++;
				}
			}

			if (ccc == '\r') {						/* CR pressed, then process the command			*/
				Cmd[g_CmdIdx] = '\0';
				TaskIdx = -1;
				RdIdx   = 0;
				while ((Cmd[RdIdx] == ' ')			/* Skip start of line white spaces				*/
				||     (Cmd[RdIdx] == ',')
				||     (Cmd[RdIdx] == '\t')) {
					RdIdx++;
				}

				Done  = 0;
				Error = 0;

				if (g_CmdIdx == RdIdx) {			/* When everything white space, nothing to do	*/
					Done = 1;
				}

				g_CmdIdx = 0;

				printf("\033[15;0H");

				if ((0 == strncmp(&Cmd[RdIdx], "?", 1))
				||  (0 == strncmp(&Cmd[RdIdx], "INFO", 4))) {
					putchar('\n');
					putchar('\n');
					for (ii=0 ; ii<sizeof(TskDsc)/sizeof(TskDsc[0]) ; ii++) {
						printf("%s  -->  ",  TskDsc[ii].Task->TskName);
					  #if ((OX_ROUND_ROBIN) != 0)
						printf("RR %3d  |  ", TSKgetRR(TskDsc[ii].Task));
					  #endif
					  #if ((OX_STARVE_RUN_MAX) != 0)
						printf("SP %1d  |  ", TSKgetStrvPrio(TskDsc[ii].Task));
						printf("SR %3d  |  ", TSKgetStrvRunMax(TskDsc[ii].Task));
						printf("SW %3d  |  ", TSKgetStrvWaitMax(TskDsc[ii].Task));
					  #endif
						printf("P %d      \n",  TskDsc[ii].Task->OriPrio);
					}
					putchar('\n');
					Done = 1;
					printf("\033[15;15H       ");
				}

				if (0 == strncmp(&Cmd[RdIdx], "HI", 2)) {
					TaskIdx = TSK_HIGH;
					RdIdx += 2;
				}
				if (0 == strncmp(&Cmd[RdIdx], "RRA", 3)) {
					TaskIdx = TSK_RRA;
					RdIdx += 3;
				}
				if (0 == strncmp(&Cmd[RdIdx], "RRB", 3)) {
					TaskIdx = TSK_RRB;
					RdIdx += 3;
				}
				if (0 == strncmp(&Cmd[RdIdx], "RRC", 3)) {
					TaskIdx = TSK_RRC;
					RdIdx += 3;
				}
				if (0 == strncmp(&Cmd[RdIdx], "STRVA", 5)) {
					TaskIdx = TSK_STRVA;
					RdIdx += 5;
				}
				if (0 == strncmp(&Cmd[RdIdx], "STRVB", 5)) {
					TaskIdx = TSK_STRVB;
					RdIdx += 5;
				}
				if (Done == 0) {
					if (TaskIdx < 0) {
						printf("\033[15;15H");		/* Set cursor after Abassi Demo >				*/
						ii = printf("Unrecognized task name");
						for( ; ii<sizeof(Cmd) ; ii++) {
							putchar(' ');
						}
						Error=1;
					}
				}

				Op = -1;							/* 2nd token if the service, check all			*/
				if ((Error == 0)
				&&  (Done == 0)) {
					while ((Cmd[RdIdx] == ' ')		/* Skip white spaces between 1st and 2nd token	*/
					||     (Cmd[RdIdx] == ',')
					||     (Cmd[RdIdx] == '\t')) {
						RdIdx++;
					}
					if (0 == strncmp(&Cmd[RdIdx], "RR", 2)) {
						Op = OP_ROUND_ROBIN;		/* Token names are re-mapped to operations #	*/
						RdIdx += 2;
					}
					if (0 == strncmp(&Cmd[RdIdx], "SR", 2)) {
						Op = OP_STARVE_RUN;
						RdIdx += 2;
					}
					if (0 == strncmp(&Cmd[RdIdx], "SP", 2)) {
						Op = OP_STARVE_PRIO;
						RdIdx += 2;
					}
					if (0 == strncmp(&Cmd[RdIdx], "SW", 2)) {
						Op = OP_STARVE_WAIT;
						RdIdx += 2;
					}
					if (0 == strncmp(&Cmd[RdIdx], "P", 1)) {
						Op = OP_PRIO;
						RdIdx += 1;
					}
					if (Op < 0) {
						Error = 1;
						printf("\033[15;15H");		/* Set cursor after Abassi Demo >				*/
						ii = printf("Unrecognized property        ");
						for( ; ii<sizeof(Cmd) ; ii++) {
							putchar(' ');
						}
					}
				}

				if ((Error == 0)
				&&  (Done == 0)) {
					while ((Cmd[RdIdx] == ' ')		/* Skip white spaces between 2nd and 3rd token	*/
					||     (Cmd[RdIdx] == ',')
					||     (Cmd[RdIdx] == '\t')) {
						RdIdx++;
					}
					ii = 0;
					if (Cmd[RdIdx] == '\0') {
						Error = 1;
						printf("\033[15;15H");		/* Set cursor after mAbassi Demo>				*/
						ii = printf("Invalid number");
						for( ; ii<sizeof(Cmd) ; ii++) {
							putchar(' ');
						}
					}
					do {							/* 3rd token is the service numerical value		*/
						if ((Cmd[RdIdx] >= '0')
						&&  (Cmd[RdIdx] <= '9')) {
							ii = ii*10;
							ii += Cmd[RdIdx] - '0';
							RdIdx ++;
						}
						else if (Cmd[RdIdx] == '\0') {
							Done = 1;
						}
						else {
							Error = 1;
							printf("\033[15;15H");	/* Set cursor after Abassi Demo >				*/
							ii = printf("Invalid number");
							for( ; ii<sizeof(Cmd) ; ii++) {
								putchar(' ');
							}
						}
					} while ((Error == 0)
					  &&     (Done  == 0));

					if (Error == 0) {				/* When everything OK, set the numerical value	*/
						switch(Op) {				/* to the service in the task					*/
						case OP_ROUND_ROBIN:
						  #if ((OX_ROUND_ROBIN) < 0)
							TSKsetRR(TskDsc[TaskIdx].Task, ii);
						  #endif
							break;
						case OP_STARVE_PRIO:
						  #if ((OX_STARVE_PRIO) < 0)
							TSKsetStrvPrio(TskDsc[TaskIdx].Task, ii);
						  #endif
							break;
						case OP_STARVE_RUN:
						  #if ((OX_STARVE_RUN_MAX) < 0)
							TSKsetStrvRunMax(TskDsc[TaskIdx].Task, ii);
						  #endif
							break;
						case OP_STARVE_WAIT:
						  #if ((OX_STARVE_WAIT_MAX) < 0)
							TSKsetStrvWaitMax(TskDsc[TaskIdx].Task, ii);
						  #endif
							break;
						default:
							TSKsetPrio(TskDsc[TaskIdx].Task, ii);
							break;
						}
					}
				}
				if (Error == 0) {
					printf("\033[15;0H");			/* Set cursor at beginning of line				*/
					printf("Abassi Demo > ");
					for (ii=0 ; ii<sizeof(Cmd) ; ii++) {
						putchar(' ');
					}
				}
				printf("\033[15;15H");
			}
		}
	}
}

/* ------------------------------------------------------------------------------------------------ */
/*																									*/
/* Idle Task																						*/
/*																									*/
/* ------------------------------------------------------------------------------------------------ */

void IdleTask(void)
{
DispInfo_t *Disp;
int         ii;
intptr_t    Msg;

	MbxDisp = MBXopen("Disp", MBX_DISP_SIZE);		/* Example of run time safe creation			*/

	for(;;) {
		ii = OSdint();								/* Will print, don't want to be pre-empted		*/
		if (0 ==  MBXget(MbxDisp, &Msg, 0)) {		/* Check if A&E has sent a string to display	*/
			Disp = (void *)Msg;						/* on the LCD									*/
			LCD_Write(Disp->Str,  Disp->Y, Disp->X);/* of updaqting the LCD							*/
		  	LCD_Update(1, 0, Disp->X);
		}
		OSeint(ii);
	}
}

/* ------------------------------------------------------------------------------------------------ */
/*																									*/
/* Tasks function																					*/
/*																									*/
/* All tasks use the same function.																	*/
/* The Task Argument is the custom task description data structure									*/
/*																									*/
/* ------------------------------------------------------------------------------------------------ */

void Fct(void)
{
DispInfo_t *Disp;
int         ii;
intptr_t    Msg;
int         NoNewMsg;
TskDsc_t   *MyInfo;


	MyInfo = (TskDsc_t *)TSKgetArg();				/* Get the argument for this task				*/

	for ( ; ; ) {
	  #if USE_EVENTS								/* Info from ISR state machine comes from 		*/
		NoNewMsg = EVTwait(0x0000, 0xFFFF, 0);		/* either an event flag or a mailbox			*/
		Msg      = EVTget();
	  #else
		NoNewMsg = MBXget(MyInfo->Mailbox, &Msg, 0);
	  #endif

		if (NoNewMsg == 0) {						/* When a valid sync trigger was received		*/

			if ((int)Msg & OP_ZERO_CNT) {			/* Request to zero the running time				*/
				MyInfo->CPUcnt   = 0;
				MyInfo->WasReset = 0;
			}
			if (((int)Msg & OP_LOCK_MTX)			/* Request from ISR to lock the mutex			*/
			&&  (MTXowner(Mutex) != TSKmyID())) {	/* Safety in case state machine has issues		*/
				MyInfo->WillBlk = 0;
				MyInfo->WillMtx = 0;
				MTXlock(Mutex, -1);
			}
			if (((int)Msg & OP_UNLOCK_MTX)			/* Request from ISR to unlock the mutex			*/
			&&  (MTXowner(Mutex) == TSKmyID())) {	/* Safety in case state machine has issues		*/
				MTXunlock(Mutex);
			}
		}
		ii = OSdint();								/* Will print, don't want to be pre-empted		*/
		if (0 ==  MBXget(MbxDisp, &Msg, 0)) {		/* Check if A&E has sent a string to display	*/
			Disp = (void *)Msg;						/* on the LCD									*/
			LCD_Write(Disp->Str,  Disp->Y, Disp->X);/* of updaqting the LCD							*/
		  	LCD_Update(1, 0, Disp->X);
		}
		OSeint(ii);
	}
}

/* ------------------------------------------------------------------------------------------------ */
/*																									*/
/* RTOS Timer call back function																	*/
/*																									*/
/* ------------------------------------------------------------------------------------------------ */

#if USE_EVENTS
	#define SEND_INFO(_Idx, _Info)	EVTset(TskDsc[_Idx].Task, _Info);
#else
	#define SEND_INFO(_Idx, _Info)	MBXput(TskDsc[_Idx].Mailbox, _Info, 0)
#endif

void TIMcallBack(void)
{
int    AErunning;
int    ii;
int    jj;
int    RunIdx;
static int Cnt       =  0;
static int MtxCnt    =  0;
static int MtxOwner  =  5;
static int MtxSelect =  5;
static int MtxBlk    =  1;
static int MtxState  =  0;
static int SuspCnt   =  0;

	Cnt++;											/* Check if A&E is running as we can't update	*/
	AErunning = (G_OStaskNow == AdamEve);			/* everything when it does						*/
													/* Update the CPU of the tasks					*/
	RunIdx = -1;									/* Invalidate running status of the  task		*/
	if (AErunning == 0) {							/* When A&E is running, don't update			*/
		for (ii=0 ; ii<sizeof(TskDsc)/sizeof(TskDsc[0]) ; ii++) {
			TskDsc[ii].Running = 0;
			if (G_OStaskNow == TskDsc[ii].Task) {
				TskDsc[ii].Running  = 1;
				RunIdx = ii;
			}
			if (TskDsc[ii].Task->Blocker != (SEM_t *)NULL) {
				if (TskDsc[ii].WasReset == 0) {
					SEND_INFO(ii, OP_ZERO_CNT);
					TskDsc[ii].WasReset = 1;
				}
			}
			if (TskDsc[ii].Task->RRmax != 0U) {
				if (TskDsc[ii].Task->RRcnt <= 1U) {
					if (TskDsc[ii].WasReset == 0) {
						SEND_INFO(ii, OP_ZERO_CNT);
						TskDsc[ii].WasReset = 1;
					}
				}
			}
		}
	}

	for (ii=0 ; ii<sizeof(TskDsc)/sizeof(TskDsc[0]) ; ii++) {	/* Excluding A&E & Idles			*/
		if (TskDsc[ii].Running != 0) {				/* It is supposed to be running, check if so	*/
			TskDsc[ii].CPUcnt++;
		}
	}

	if (AErunning == 0) {
		switch(MtxState) {
		case 0:										/* State #0 : new mutex owner					*/
			if (++MtxSelect >= sizeof(TskDsc)/sizeof(TskDsc[0])) {
				MtxSelect = 0;
			}
			if (MtxSelect == 2) {					/* RRB if not the first locker of the mutex		*/
				MtxSelect = 4;
			}

			MtxOwner = MtxSelect;

			SEND_INFO(MtxOwner, OP_LOCK_MTX);		/* Tell new owner to lock the mutex				*/

			if (++MtxBlk >= 4) {					/* Select the task that will block				*/
				MtxBlk = 0;							/* Only Hi & RR can block on the mutex			*/
			}

			if (MtxBlk == MtxOwner) {				/* Get another task to block on the mutex		*/
				if (++MtxBlk >= 4) {
					MtxBlk = 0;
				}
			}
			TskDsc[MtxOwner].WillMtx = 1;			/* Display it will lock the mutex soon			*/
			TskDsc[MtxBlk].WillBlk = 1;				/* And this will be blocked 					*/
			MtxState++;
			break;
		case 1:										/* Waiting for the owner to lock the mutex		*/
			if (TskDsc[MtxOwner].Task->MyMutex == Mutex) {
				MtxState++;
			}
			break;
		case 2:										/* OK, the mutex owner is locking the mutex		*/
			SEND_INFO(MtxBlk, OP_LOCK_MTX);			/* Tell the task to be blocked to lock it		*/
			MtxState++;
			break;
		case 3:										/* Make sure the task gets blocked				*/
			if (TskDsc[MtxOwner].Task->Prio == TskDsc[MtxBlk].Task->Prio) {
				if (TSKisBlk(TskDsc[MtxBlk].Task)) {/* OK, the task is blocked on the mutex now		*/
					MtxState++;						/* This may take a while as it needs to run		*/
					MtxCnt = 0;
				}
			}
			else {									/* If to be suspended, must skip a few states	*/
				if (TskDsc[MtxOwner].Task->SuspRqst != 0) {	/* as != prio can't run at same time	*/
					MtxState = 4;
					MtxCnt   = 0;
					TskDsc[MtxOwner].WillMtx = 0;
					TskDsc[MtxBlk].WillBlk   = 0;
					TskDsc[MtxBlk].WillMtx   = 1;
				}
			}
			break;
		case 4:										/* State #4: Task is blocked on the mutex		*/
			if (MtxOwner == RunIdx) {				/* Make sure the owner is running long enough	*/
				jj = (MtxOwner == 0)				/* When HI task, is getting suspended			*/
				   ? (((15*OS_TICK_PER_SEC)+1) - TskDsc[0].CPUcnt)
				   :   ((5*OS_TICK_PER_SEC)+1);
				if (++MtxCnt > jj) {				/* Tell the owner to release the mutex			*/
					SEND_INFO(MtxOwner, OP_UNLOCK_MTX);
					MtxState++;
				}
			}
			break;
		case 5:										/* Wait for owner to unlock the mutex			*/
			if (TskDsc[MtxOwner].Task->MyMutex != Mutex) {
				MtxOwner = MtxBlk;					/* The new owner is the one that was blocked	*/
				MtxState++;
			}
			break;
		case 6:										/* Wait for new owner to lock the mutex			*/
			if (TskDsc[MtxOwner].Task->MyMutex == Mutex) {
				MtxState++;
				MtxCnt = 0;
			}
			break;
		case 7:
			if (MtxOwner == RunIdx) {				/* Wait a while before telling to release		*/
				jj = (MtxOwner == 0)				/* When HI task, is getting suspended			*/
				   ? (((15*OS_TICK_PER_SEC)+1) - TskDsc[0].CPUcnt)
				   :   ((5*OS_TICK_PER_SEC)+1);
				if (++MtxCnt > jj) {				/* Tell the owner to release the mutex			*/
					SEND_INFO(MtxOwner, OP_UNLOCK_MTX);
					MtxState++;
				}
			}
			break;
		default:									/* Wait for blocked to unlock					*/
			if (TskDsc[MtxOwner].Task->MyMutex != Mutex) {
				MtxState =  0;
				MtxOwner = -1;
			}
			break;
  		}

		if ((Cnt >= OS_TICK_PER_SEC/TSK_REFRESH_RATE)
		&&  (TSKisBlk(AdamEve))) {					/* Make sure A&E does not use all the CPU		*/
			SEMpost(SemaAE);						/* Tell A & E to refresh the display			*/
			Cnt = 0;
		}
	}

	SuspCnt++;
	if (SuspCnt == 10*OS_TICK_PER_SEC+3) {			/* Suspend / resume the high priority			*/
		TSKsuspend(TskDsc[TSK_HIGH].Task);
	}
	else if (SuspCnt > 30*OS_TICK_PER_SEC+3) {
		TSKresume(TskDsc[TSK_HIGH].Task);
		SuspCnt = 0;
	}

	return;
}

/* ------------------------------------------------------------------------------------------------ */
/*																									*/
/* Initialization functions																			*/
/*																									*/
/* ------------------------------------------------------------------------------------------------ */

void CLOCK_XT2_Init()
{
int timeout;

	timeout = 5000;

	P11DIR  = 0x01
		    | 0x02
	        | 0x04;									/* set P11.0, P11.1 & P11.2 as output			*/
	P11SEL  = 0x01
		    | 0x02
		    | 0x04;									/* set P11.0/.1/.2 as ACLK/MCLK/SMCLK function	*/
	P5SEL   = P5SEL
		    | 0x04
		    | 0x08;									/* set P5.2 & P5.3 as XT2IN/XT2OUT				*/
	UCSCTL6 = 0x80									/* XT2 drive for 16 -> 24 MHz					*/
	        | 0x10;									/* XT2 soruce extyernally from the pin			*/
	UCSCTL3 = UCSCTL3 | 0x20;						/* set SELREF = 010 --> FLL clock src = REFOCLK	*/
	UCSCTL4	= 0x0005								/*	MCLK source == XT2CLK						*/
			| 0x0050								/* SMCLK source == XT2CLK						*/
			| 0x0500;								/*	ACLK source == XT2CLK						*/
	UCSCTL5 = 0x0001								/* MCLK : /2 									*/
	        | 0x0004								/* SMCLK: /16									*/
	        | 0x0100;								/* ACLK : /2									*/
	do {
		if (!(UCSCTL7 & 0x08)) break;
		UCSCTL7 = UCSCTL7 & ~(0x0F);				/* Clear XT2,XT1,DCO fault flags				*/
		SFRIFG1 = SFRIFG1 & ~(0x02);				/* Clear fault flags (SFRIFG1.OFIFG = 0)		*/
		timeout--;
	} while ((SFRIFG1 & 0x02) && (timeout != 0));	/* Test oscillator fault flag					*/

	return;
}

/* ------------------------------------------------------------------------------------------------ */

void TIMERinit(long Rate)
{
int PreSc;

	for (PreSc=0 ; PreSc<3 ; PreSc++) {				/* Find ther pre-scaling to use to have a count	*/
		if (Rate <= 65535L) {						/* held on 16 bits								*/
			break;
		}
		Rate = Rate >> 1;
	}

	if (Rate > 65535L) {
		Rate = 65535L;
	}
	else {
		Rate--;
	}

	TA0CTL	 = 0x0100								/* Timer A clock source is ACLK					*/
			 | (PreSc << 6)							/* Divider selection							*/
			 | 0x0010;								/* Mode up count								*/
	TA0CCR0	 = (int)Rate;
	TA0CCTL0 = CCIE;								/* Enable CCR0 interrupt of timer A0			*/

	return;
}

/* ------------------------------------------------------------------------------------------------ */

void SPI2_Init(void)
{
unsigned int prescaler;

	P9SEL    |= 0x0E;								/* P9.1/P9.2/P9.3 set as alternate function		*/
	P9DIR    |= 0x0B;								/* P9.0/P9.1/P9.3 set as output					*/
	P9DIR    |= 0xC0;								/* LCD Reset, data/command bits set as output	*/
	
	UCB2CTL1 |= 0x01;								/* UCSWRST = 1 (reset state enable)				*/

	/* set format:																					*/
	/* set clock polarity and phase (UCCKPH = 0; UCCKPL = 1) --> bit7=0; bit6=1						*/
	/* MSB first; bit0 (UCSYNC) = 1; bit3 (master mode) = 1; --> bit5=1; bit4=0; bit3 =3; bit0=1;	*/
	/* set mode: 3pin SPI --> bit1=0; bit2=3;														*/
	/* UCB2CTL0 = 0b01101001 = 0x69																	*/

	UCB2CTL0  = 0x69;

	UCB2CTL1 |= 0x80;								/* Clock source - SMCLK							*/

  #if ((OSC_FREQ / LCD_FREQ) / 16) == 0
	prescaler = 1;
  #else
	prescaler = ((OSC_FREQ / LCD_FREQ) / 16);
  #endif
	UCB2BR0   = (prescaler &  0xff);
	UCB2BR1   = 0xFF & (prescaler >> 8);

	P9SEL    |= 0x0E;				  				/* P9.0/P9.1/P9.2/P9.3 set as alternate fct		*/
	P9DIR    |= 0x0B;								/* P9.0/P9.1/P9.3 set as output					*/

	UCB2CTL1 &=  ~(0x01);							/* UCSWRST = 0 (reset state disable)			*/

	return;

}

/* ------------------------------------------------------------------------------------------------ */
/* LCD functions																					*/
/* ------------------------------------------------------------------------------------------------ */

void LCD_Init()
{
	P9OUT = P9OUT & 0xBF;
	Delay(10000);
	P9OUT = P9OUT | 0x40;
	Delay(10000);


	LCD_Send(0x21, SEND_CMD);
	LCD_Send(0xC8, SEND_CMD);
	LCD_Send(0x04 | (!!(LCD_START_LINE_ADDR&(0x40))),   SEND_CMD);
	LCD_Send(0x40 | (LCD_START_LINE_ADDR & ((0x40)-1)), SEND_CMD);
	LCD_Send(0x12, SEND_CMD);
	LCD_Send(0x20, SEND_CMD);
	LCD_Send(0x08, SEND_CMD);
	LCD_Send(0x0C, SEND_CMD);

	LCD_Contrast(0xFF);
	LCD_Clear();
	LCD_Write(" ", 0, 0);

	return;
}

/* ------------------------------------------------------------------------------------------------ */

void LCD_Send(int data, int cmd)
{
	P9OUT = P9OUT & ~(0x01);

	if (cmd == SEND_CHR) {
		P9OUT |= 0x80;								/* Send char									*/
	}
	else {
		P9OUT &= 0x7F;								/* Send command									*/
	}

	while (!(UCB2IFG & 0x02));						/* Transmit the data over the SPI				*/
	UCB2TXBUF = data;

	P9OUT = P9OUT | 0x01;

	return;
}

/* ------------------------------------------------------------------------------------------------ */

void LCD_Clear_Memory()
{
	for (g_LCDidx=0; g_LCDidx<504; g_LCDidx++) {
		LCDMemory[g_LCDidx] = 0;
	}
	g_LCDidx = 0;

	return;
}

/* ------------------------------------------------------------------------------------------------ */

void LCD_Update(int mode, int num, int y)
{
char x;

	LCD_Send(0x80,	 SEND_CMD);
	LCD_Send(0x40|y, SEND_CMD);

	if (((mode == 1) && !(num & 1<<y))
	||  ((mode != 1) &&  (num & 1<<y))) {
		for(x=0; x<84; x++) {
			LCD_Send(LCDMemory[y*84+x], SEND_CHR);
		}
	}
	else {
		for(x=0; x<84; x++) {
			LCD_Send(~LCDMemory[y*84+x], SEND_CHR);
		}
	}

	return;
}

/* ------------------------------------------------------------------------------------------------ */

void LCD_Clear()
{
int ii;

	LCD_Clear_Memory();
	for (ii=0 ; ii<6 ; ii++) {
		LCD_Update(1, 0x00, ii);
	}
	return;
}

/* ------------------------------------------------------------------------------------------------ */

void LCD_Contrast(int contrast)
{
	LCD_Send(0x21, SEND_CMD);
	LCD_Send(0x80 | contrast, SEND_CMD);
	LCD_Send(0x20, SEND_CMD);

	return;
}

/* ------------------------------------------------------------------------------------------------ */

void LCD_Write(char *sss, int x, int y)
{
char ii;

	y        = y + x/84;
	y        = y % 6;
	x        = x % 84;

	g_LCDidx = 84*y + x;

	LCD_Send(0x80 | x, SEND_CMD);
	LCD_Send(0x40 | y, SEND_CMD);

	while (*sss) {
		for (ii=0; ii<5; ii++) {					/* Charac are 5X7 pixels						*/
			if (g_LCDidx < 504) {
				LCDMemory[g_LCDidx++] = FontLookup[*sss-' '][ii];
			}
		}
		if (g_LCDidx < 504) {
			LCDMemory[g_LCDidx++] = 0x00;			/* 	space between characters					*/
		}
		sss++;
	}
	return;
}

/* ------------------------------------------------------------------------------------------------ */

void Delay(unsigned long time)
{

	while (time--);
}

/* ------------------------------------------------------------------------------------------------ */
/*																									*/
/* UART functions needed by the Peripheral library to access the UART								*/
/*																									*/
/* ------------------------------------------------------------------------------------------------ */

int UARTopen(const char *path, unsigned int flags, int fd)
{
	return(0);
}

/* ------------------------------------------------------------------------------------------------ */

int UARTclose(int fd)
{
	return(0);
}

/* ------------------------------------------------------------------------------------------------ */

int UARTread(int fd, char *buf, unsigned int count)
{
	*buf = __getchar();
	return(1);
}

/* ------------------------------------------------------------------------------------------------ */

int UARTwrite(int fd, const char *buf, unsigned int count)
{
int ii;
	for(ii=0 ; ii<count ; ii++) {
		 __putchar(*buf++);
	}
	return(count);
}

/* ------------------------------------------------------------------------------------------------ */

off_t UARTseek(int fd, off_t off, int start)
{
	return(0);
}

/* ------------------------------------------------------------------------------------------------ */

int UARTunlink(const char *path)
{
	return(0);
}

/* ------------------------------------------------------------------------------------------------ */

int UARTrename(const char *old, const char *new)
{
	return(0);
}

/* ------------------------------------------------------------------------------------------------ */

void UARTinit(void)
{

	UART2_Init(BAUDRATE);
	UART3_Init(BAUDRATE);

	add_device("uart", _MSA, UARTopen, UARTclose, UARTread, UARTwrite, UARTseek, UARTunlink, UARTrename);
	freopen("uart:stdout", "w", stdout);
	setvbuf(stdout, NULL, _IONBF, 0);
	freopen("uart:stdin", "r", stdin);
	setvbuf(stdin, NULL, _IONBF, 0);

	return;
}

/* ------------------------------------------------------------------------------------------------ */

void UART2_Init(unsigned long baud)
{
float n, x;

	P10SEL    = 0;
	P10DIR   |= (1<<7)								/* set P10.7 as output LED 1					*/
	         |  (1<<6);								/* set P10.6 as output LED 2					*/
	P9DIR    |= 1<<4;								/* set P9.4=1 (Transmitter) as output			*/
	P9DIR    &= ~(1<<5);							/* set P9.5=0 (Receiver) as input				*/
	P9SEL    |= (1<<4)								/* set P9.4=1 & P9.5=1 as TX/RX					*/
	         |  (1<<5);
	UCA2CTL1 |= 0x01;								/* UCSWRST = 1 (bit0 = 1)						*/
	UCA2CTL1 = (UCA2CTL1 | (1<<6) | 0x01)			/* select ACLK as BaudRate source clock			*/
	         & ~(1<<7);								/* (bit6 = 1; bit7 = 1) 						*/
	UCA2MCTL |= 0x01;								/* UCOS16 = 1									*/
	n        = SYS_FREQ / baud;						/* calculate UCBR								*/
	UCA2BR0  = (int)(n/16) % 256;
	UCA2BR1  = (int)(n/16) / 256;
	x        = (int)((n-(int)(n))*8+0.5);			/* caluclate UCBRS								*/
	UCA2MCTL = UCA2MCTL | (int)(x)<<1;
	x        = (int)(((n/16 - (int)(n/16))*16) + 0.5);
	UCA2MCTL = UCA2MCTL | (int)(x)<<4;				/* caluclate UCBRF								*/
													/* calculate registers for the BaudRate			*/
	UCA2CTL0 = 0x00;								/* Set Format Frame: 8 bit; no parity; 1 stop	*/
	UCA2CTL1 = UCA2CTL1 & ~(0x01);					/* UCSWRST = 0 (bit0 = 0)						*/

	return;
}

/* ------------------------------------------------------------------------------------------------ */

void UART3_Init(unsigned long baud)
{
float n, x;

	P10DIR   |= 1<<4;								/* set P9.4=1 (Transmitter) as output			*/
	P10DIR   &= ~(1<<5);							/* set P9.5=0 (Receiver) as input				*/
	P10SEL   |= (1<<4)								/* set P9.4=1 & P9.5=1 as TX/RX					*/
	         |  (1<<5);
	UCA3CTL1 |= 0x01;								/* UCSWRST = 1 (bit0 = 1)						*/
	UCA3CTL1 = (UCA3CTL1 | (1<<6) | 0x01)			/* select ACLK as BaudRate source clock			*/
	         & ~(1<<7);								/* (bit6 = 1; bit7 = 1) 						*/
	UCA3MCTL |= 0x01;								/* UCOS16 = 1									*/
	n        = SYS_FREQ / baud;						/* calculate UCBR								*/
	UCA3BR0  = (int)(n/16) % 256;
	UCA3BR1  = (int)(n/16) / 256;
	x        = (int)((n-(int)(n))*8+0.5);			/* caluclate UCBRS								*/
	UCA3MCTL = UCA3MCTL | (int)(x)<<1;
	x        = (int)(((n/16 - (int)(n/16))*16) + 0.5);
	UCA3MCTL = UCA3MCTL | (int)(x)<<4;				/* caluclate UCBRF								*/
													/* calculate registers for the BaudRate			*/
	UCA3CTL0 = 0x00;								/* Set Format Frame: 8 bit; no parity; 1 stop	*/
	UCA3CTL1 = UCA3CTL1 & ~(0x01);					/* UCSWRST = 0 (bit0 = 0)						*/

	return;
}

/* ------------------------------------------------------------------------------------------------ */

int __putchar(int ccc)
{
	if (ccc == '\n') {
		while((UCA2IFG & 0x02) == 0);
		UCA2TXBUF = '\r';
		while((UCA3IFG & 0x02) == 0);
		UCA3TXBUF = '\r';
	}
	while (!(UCA2IFG & 0x02));
	UCA2TXBUF = ccc;
	while (!(UCA3IFG & 0x02));
	UCA3TXBUF = ccc;

	return(ccc);
}

/* ------------------------------------------------------------------------------------------------ */

int __getchar(void)
{

	while (((UCA2IFG & 0x01) == 0)
	&&     ((UCA3IFG & 0x01) == 0));

	if ((UCA2IFG & 0x01) != 0){
		return(UCA2RXBUF);
	}
	return(UCA3RXBUF);
}

/* ------------------------------------------------------------------------------------------------ */

int GetKey(void)
{
int ccc;

	ccc = 0;

	if ((UCA2IFG & 0x01) != 0) {					/* Check both serial ports						*/
		ccc = UCA2RXBUF;
	}
	if ((ccc == 0)
	&&  ((UCA3IFG & 0x01) != 0)) {
		ccc = UCA3RXBUF;
	}

	return(ccc);
}

/* EOF */
